/* Author: Mark Goldman   <mgoldman@tresys.com>
 * 			Paul Rosenfeld	<prosenfeld@tresys.com>
 *
 * Copyright (C) 2007 Tresys Technology, LLC
 *
 *  This library is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as
 *  published by the Free Software Foundation; either version 2.1 of the
 *  License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 *  02110-1301  USA
 */

#include <semanage/handle.h>
#include <semanage/seusers_policy.h>
#include <semanage/users_policy.h>
#include <semanage/user_record.h>
#include "semanage_store.h"
#include "seuser_internal.h"
#include "debug.h"

#include "utilities.h"
#include "genhomedircon.h"
#include <ustr.h>

#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <unistd.h>

/* paths used in get_home_dirs() */
#define PATH_ETC_USERADD "/etc/default/useradd"
#define PATH_ETC_LIBUSER "/etc/libuser.conf"
#define PATH_DEFAULT_HOME "/home"
#define PATH_EXPORT_HOME "/export/home"
#define PATH_ETC_LOGIN_DEFS "/etc/login.defs"

/* other paths */
#define PATH_SHELLS_FILE "/etc/shells"
#define PATH_NOLOGIN_SHELL "/sbin/nologin"

/* comments written to context file */
#define COMMENT_FILE_CONTEXT_HEADER "#\n#\n# " \
			"User-specific file contexts, generated via libsemanage\n" \
			"# use semanage command to manage system users to change" \
			" the file_context\n#\n#\n"

#define COMMENT_USER_HOME_CONTEXT "\n\n#\n# Home Context for user %s" \
			"\n#\n\n"

/* placeholders used in the template file
   which are searched for and replaced */
#define TEMPLATE_HOME_ROOT "HOME_ROOT"
#define TEMPLATE_HOME_DIR "HOME_DIR"
#define TEMPLATE_USER "USER"
#define TEMPLATE_ROLE "ROLE"
#define TEMPLATE_SEUSER "system_u"

#define FALLBACK_USER "user_u"
#define FALLBACK_USER_PREFIX "user"
#define DEFAULT_LOGIN "__default__"

typedef struct {
	const char *fcfilepath;
	int usepasswd;
	const char *homedir_template_path;
	semanage_handle_t *h_semanage;
} genhomedircon_settings_t;

typedef struct user_entry {
	char *name;
	char *sename;
	char *prefix;
	char *home;
	struct user_entry *next;
} genhomedircon_user_entry_t;

typedef struct {
	const char *search_for;
	const char *replace_with;
} replacement_pair_t;

static semanage_list_t *default_shell_list(void)
{
	semanage_list_t *list = NULL;

	if (semanage_list_push(&list, "/bin/csh")
	    || semanage_list_push(&list, "/bin/tcsh")
	    || semanage_list_push(&list, "/bin/ksh")
	    || semanage_list_push(&list, "/bin/bsh")
	    || semanage_list_push(&list, "/bin/ash")
	    || semanage_list_push(&list, "/usr/bin/ksh")
	    || semanage_list_push(&list, "/usr/bin/pdksh")
	    || semanage_list_push(&list, "/bin/zsh")
	    || semanage_list_push(&list, "/bin/sh")
	    || semanage_list_push(&list, "/bin/bash"))
		goto fail;

	return list;

      fail:
	semanage_list_destroy(&list);
	return NULL;
}

static semanage_list_t *get_shell_list(void)
{
	FILE *shells;
	char *temp = NULL;
	semanage_list_t *list = NULL;
	size_t buff_len = 0;

	shells = fopen(PATH_SHELLS_FILE, "r");
	if (!shells)
		return default_shell_list();
	while (getline(&temp, &buff_len, shells) >= 0) {
		if (strcmp(temp, PATH_NOLOGIN_SHELL)) {
			if (semanage_list_push(&list, temp)) {
				free(temp);
				semanage_list_destroy(&list);
				return default_shell_list();
			}
		}
	}
	free(temp);

	return list;
}

static semanage_list_t *get_home_dirs(genhomedircon_settings_t * s)
{
	semanage_list_t *homedir_list = NULL;
	semanage_list_t *shells = NULL;
	char *rbuf = NULL;
	char *path = NULL;
	long rbuflen;
	size_t minuid = 0;
	size_t minuid_set = 0;
	size_t temp;
	struct passwd pwstorage, *pwbuf;
	struct stat buf;

	shells = get_shell_list();
	assert(shells);

	path = semanage_findval(PATH_ETC_USERADD, "HOME", "=");
	if (path && *path) {
		if (semanage_list_push(&homedir_list, path)) {
			free(path);
			goto fail;
		}
	}
	free(path);

	path = semanage_findval(PATH_ETC_LIBUSER, "LU_HOMEDIRECTORY", "=");
	if (path && *path) {
		if (semanage_list_push(&homedir_list, path)) {
			free(path);
			goto fail;
		}
	}
	free(path);

	if (!homedir_list) {
		if (semanage_list_push(&homedir_list, PATH_DEFAULT_HOME)) {
			goto fail;
		}
	}

	if (!stat(PATH_EXPORT_HOME, &buf)) {
		if (S_ISDIR(buf.st_mode)) {
			if (semanage_list_push(&homedir_list, PATH_EXPORT_HOME)) {
				goto fail;
			}
		}
	}

	if (!(s->usepasswd))
		return homedir_list;

	path = semanage_findval(PATH_ETC_LOGIN_DEFS, "UID_MIN", NULL);
	if (path && *path) {
		temp = atoi(path);
		if (!minuid_set || temp < minuid) {
			minuid = temp;
			minuid_set = 1;
		}
	}
	free(path);

	path = semanage_findval(PATH_ETC_LIBUSER, "LU_UIDNUMBER", "=");
	if (path && *path) {
		temp = atoi(path);
		if (!minuid_set || temp < minuid) {
			minuid = temp;
			minuid_set = 1;
		}
	}
	free(path);

	if (!minuid_set) {
		minuid = 500;
		minuid_set = 1;
	}

	rbuflen = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (rbuflen <= 0)
		goto fail;
	rbuf = malloc(rbuflen);
	if (rbuf == NULL)
		goto fail;
	setpwent();
	for (errno = 0; getpwent_r(&pwstorage, rbuf, rbuflen, &pwbuf) == 0; errno = 0) {
		if (pwbuf->pw_uid < minuid)
			continue;
		if (!semanage_list_find(shells, pwbuf->pw_shell))
			continue;
		if (strcmp(pwbuf->pw_dir, "/") == 0)
			continue;
		if (semanage_str_count(pwbuf->pw_dir, '/') <= 1)
			continue;
		if (!(path = strdup(pwbuf->pw_dir))) {
			break;
		}

		semanage_rtrim(path, '/');
		if (!semanage_list_find(homedir_list, path)) {
			if (semanage_list_push(&homedir_list, path)) {
				free(path);
				goto fail;
			}
		}
		free(path);
	}

	if (errno) {
		WARN(s->h_semanage, "Error while fetching users.  "
		     "Returning list so far.");
	}
	endpwent();
	free(rbuf);
	semanage_list_destroy(&shells);
	if (semanage_list_sort(&homedir_list))
		goto fail;

	return homedir_list;

      fail:
	endpwent();
	free(rbuf);
	semanage_list_destroy(&homedir_list);
	semanage_list_destroy(&shells);
	return NULL;
}

/**
 * @param	s	settings structure, stores various paths etc. Must never be NULL
 * @param	out	the FILE to put all the output in.
 * @return	0 on success
 */
static int write_file_context_header(genhomedircon_settings_t * s, FILE * out)
{
	if (fprintf(out, COMMENT_FILE_CONTEXT_HEADER) < 0) {
		return STATUS_ERR;
	}

	return STATUS_SUCCESS;
}

/* Predicates for use with semanage_slurp_file_filter() the homedir_template
 * file currently contains lines that serve as the template for a user's
 * homedir.
 *
 * It also contains lines that are the template for the parent of a
 * user's home directory.
 *
 * Currently, the only lines that apply to the the root of a user's home
 * directory are all prefixed with the string "HOME_ROOT".  All other
 * lines apply to a user's home directory.  If this changes the
 * following predicates need to change to reflect that.
 */
static int HOME_ROOT_PRED(const char *string)
{
	return semanage_is_prefix(string, TEMPLATE_HOME_ROOT);
}

static int HOME_DIR_PRED(const char *string)
{
	return semanage_is_prefix(string, TEMPLATE_HOME_DIR);
}

static int USER_CONTEXT_PRED(const char *string)
{
	return (int)(strstr(string, TEMPLATE_USER) != NULL);
}

/* make_tempate
 * @param	s	  the settings holding the paths to various files
 * @param	pred	function pointer to function to use as filter for slurp
 * 					file filter
 * @return   a list of lines from the template file with inappropriate
 *	    lines filtered out.
 */
static semanage_list_t *make_template(genhomedircon_settings_t * s,
				      int (*pred) (const char *))
{
	FILE *template_file = NULL;
	semanage_list_t *template_data = NULL;

	template_file = fopen(s->homedir_template_path, "r");
	if (!template_file)
		return NULL;
	template_data = semanage_slurp_file_filter(template_file, pred);
	fclose(template_file);

	return template_data;
}

static Ustr *replace_all(const char *str, const replacement_pair_t * repl)
{
	Ustr *retval = USTR_NULL;
	int i, num_replaced = 0;

	if (!str || !repl)
		goto done;
	if (!(retval = ustr_dup_cstr(str)))
		goto done;

	for (i = 0; repl[i].search_for; i++) {
		num_replaced += ustr_replace_cstr(&retval, repl[i].search_for,
						  repl[i].replace_with, 0);
	}
	if (!num_replaced)
		ustr_sc_free(&retval);

      done:
	return retval;
}

static int write_home_dir_context(FILE * out, semanage_list_t * tpl,
				  const char *user, const char *seuser,
				  const char *home, const char *role_prefix)
{
	replacement_pair_t repl[] = {
		{.search_for = TEMPLATE_SEUSER,.replace_with = seuser},
		{.search_for = TEMPLATE_HOME_DIR,.replace_with = home},
		{.search_for = TEMPLATE_ROLE,.replace_with = role_prefix},
		{NULL, NULL}
	};
	Ustr *line = USTR_NULL;

	if (fprintf(out, COMMENT_USER_HOME_CONTEXT, user) < 0)
		return STATUS_ERR;

	for (; tpl; tpl = tpl->next) {
		line = replace_all(tpl->data, repl);
		if (!line || !ustr_io_putfileline(&line, out))
			goto fail;
		ustr_sc_free(&line);
	}
	return STATUS_SUCCESS;

      fail:
	ustr_sc_free(&line);
	return STATUS_ERR;
}

static int write_home_root_context(FILE * out, semanage_list_t * tpl,
				   char *homedir)
{
	replacement_pair_t repl[] = {
		{.search_for = TEMPLATE_HOME_ROOT,.replace_with = homedir},
		{NULL, NULL}
	};
	Ustr *line = USTR_NULL;

	for (; tpl; tpl = tpl->next) {
		line = replace_all(tpl->data, repl);
		if (!line || !ustr_io_putfileline(&line, out))
			goto fail;
		ustr_sc_free(&line);
	}
	return STATUS_SUCCESS;

      fail:
	ustr_sc_free(&line);
	return STATUS_ERR;
}

static int write_user_context(FILE * out, semanage_list_t * tpl, char *user,
			      char *seuser, char *role_prefix)
{
	replacement_pair_t repl[] = {
		{.search_for = TEMPLATE_USER,.replace_with = user},
		{.search_for = TEMPLATE_ROLE,.replace_with = role_prefix},
		{.search_for = TEMPLATE_SEUSER,.replace_with = seuser},
		{NULL, NULL}
	};
	Ustr *line = USTR_NULL;

	for (; tpl; tpl = tpl->next) {
		line = replace_all(tpl->data, repl);
		if (!line || !ustr_io_putfileline(&line, out))
			goto fail;
		ustr_sc_free(&line);
	}
	return STATUS_SUCCESS;

      fail:
	ustr_sc_free(&line);
	return STATUS_ERR;
}

static int user_sort_func(semanage_user_t ** arg1, semanage_user_t ** arg2)
{
	return strcmp(semanage_user_get_name(*arg1),
		      semanage_user_get_name(*arg2));
}

static int name_user_cmp(char *key, semanage_user_t ** val)
{
	return strcmp(key, semanage_user_get_name(*val));
}

static int push_user_entry(genhomedircon_user_entry_t ** list, const char *n,
			   const char *sen, const char *pre, const char *h)
{
	genhomedircon_user_entry_t *temp = NULL;
	char *name = NULL;
	char *sename = NULL;
	char *prefix = NULL;
	char *home = NULL;

	temp = malloc(sizeof(genhomedircon_user_entry_t));
	if (!temp)
		goto cleanup;
	name = strdup(n);
	if (!name)
		goto cleanup;
	sename = strdup(sen);
	if (!sename)
		goto cleanup;
	prefix = strdup(pre);
	if (!prefix)
		goto cleanup;
	home = strdup(h);
	if (!home)
		goto cleanup;

	temp->name = name;
	temp->sename = sename;
	temp->prefix = prefix;
	temp->home = home;
	temp->next = (*list);
	(*list) = temp;

	return STATUS_SUCCESS;

      cleanup:
	free(name);
	free(sename);
	free(prefix);
	free(home);
	free(temp);
	return STATUS_ERR;
}

static void pop_user_entry(genhomedircon_user_entry_t ** list)
{
	genhomedircon_user_entry_t *temp;

	if (!list || !(*list))
		return;

	temp = *list;
	*list = temp->next;
	free(temp->name);
	free(temp->sename);
	free(temp->prefix);
	free(temp->home);
	free(temp);
}

static genhomedircon_user_entry_t *get_users(genhomedircon_settings_t * s,
					     int *errors)
{
	genhomedircon_user_entry_t *head = NULL;
	semanage_seuser_t **seuser_list = NULL;
	unsigned int nseusers = 0;
	semanage_user_t **user_list = NULL;
	unsigned int nusers = 0;
	semanage_user_t **u = NULL;
	const char *name = NULL;
	const char *seuname = NULL;
	const char *prefix = NULL;
	struct passwd pwstorage, *pwent = NULL;
	unsigned int i;
	long rbuflen;
	char *rbuf = NULL;
	int retval;

	*errors = 0;
	retval = semanage_seuser_list(s->h_semanage, &seuser_list, &nseusers);
	if (retval < 0 || (nseusers < 1)) {
		/* if there are no users, this function can't do any other work */
		return NULL;
	}

	if (semanage_user_list(s->h_semanage, &user_list, &nusers) < 0) {
		nusers = 0;
	}

	qsort(user_list, nusers, sizeof(semanage_user_t *),
	      (int (*)(const void *, const void *))&user_sort_func);

	/* Allocate space for the getpwnam_r buffer */
	rbuflen = sysconf(_SC_GETPW_R_SIZE_MAX);
	if (rbuflen <= 0)
		goto cleanup;
	rbuf = malloc(rbuflen);
	if (rbuf == NULL)
		goto cleanup;

	for (i = 0; i < nseusers; i++) {
		name = semanage_seuser_get_name(seuser_list[i]);
		seuname = semanage_seuser_get_sename(seuser_list[i]);

		if (strcmp(seuname, FALLBACK_USER) == 0)
			continue;
		if (strcmp(seuname, DEFAULT_LOGIN) == 0)
			continue;
		if (strcmp(seuname, TEMPLATE_SEUSER) == 0)
			continue;

		/* find the user structure given the name */
		u = bsearch(seuname, user_list, nusers, sizeof(semanage_user_t *),
			    (int (*)(const void *, const void *))
			    &name_user_cmp);
		if (u) {
			prefix = semanage_user_get_prefix(*u);
		} else {
			prefix = name;
		}

		errno = 0;
		if (getpwnam_r(name, &pwstorage, rbuf, rbuflen, &pwent) != 0) {
			if (errno != 0) {
				*errors = STATUS_ERR;
				goto cleanup;
			}
			WARN(s->h_semanage,
			     "user %s not in password file", name);
			continue;
		}

		if (strcmp(pwent->pw_dir, "/") == 0) {
			/* don't relabel / genhomdircon checked to see if root
			 * was the user and if so, set his home directory to
			 * /root */
			continue;
		}
		if (push_user_entry(&head, name, seuname,
				    prefix, pwent->pw_dir) != STATUS_SUCCESS) {
			*errors = STATUS_ERR;
			break;
		}
	}

      cleanup:
	free(rbuf);
	if (*errors) {
		for (; head; pop_user_entry(&head)) {
			/* the pop function takes care of all the cleanup
			   so the loop body is just empty */
		}
	}
	for (i = 0; i < nseusers; i++) {
		semanage_seuser_free(seuser_list[i]);
	}
	free(seuser_list);

	for (i = 0; i < nusers; i++) {
		semanage_user_free(user_list[i]);
	}
	free(user_list);

	return head;
}

static int write_gen_home_dir_context(FILE * out, genhomedircon_settings_t * s,
				      semanage_list_t * user_context_tpl,
				      semanage_list_t * homedir_context_tpl)
{
	genhomedircon_user_entry_t *users;
	int errors = 0;

	users = get_users(s, &errors);
	if (!users && errors) {
		return STATUS_ERR;
	}

	for (; users; pop_user_entry(&users)) {
		if (write_home_dir_context(out, homedir_context_tpl,
					   users->name,
					   users->sename, users->home,
					   users->prefix)) {
			return STATUS_ERR;
		}
		if (write_user_context(out, user_context_tpl, users->name,
				       users->sename, users->prefix)) {
			return STATUS_ERR;
		}
	}

	return STATUS_SUCCESS;
}

/**
 * @param	s	settings structure, stores various paths etc. Must never be NULL
 * @param	out	the FILE to put all the output in.
 * @return	0 on success
 */
static int write_context_file(genhomedircon_settings_t * s, FILE * out)
{
	semanage_list_t *homedirs = NULL;
	semanage_list_t *h = NULL;
	semanage_list_t *user_context_tpl = NULL;
	semanage_list_t *homedir_context_tpl = NULL;
	semanage_list_t *homeroot_context_tpl = NULL;
	int retval = STATUS_SUCCESS;

	homedirs = get_home_dirs(s);
	if (!homedirs) {
		WARN(s->h_semanage,
		     "no home directories were available, exiting without writing");
		return STATUS_ERR;	/* No homedirs so no output */
	}

	if (write_file_context_header(s, out) != STATUS_SUCCESS)
		return STATUS_ERR;

	homedir_context_tpl = make_template(s, &HOME_DIR_PRED);
	homeroot_context_tpl = make_template(s, &HOME_ROOT_PRED);
	user_context_tpl = make_template(s, &USER_CONTEXT_PRED);
	if (!homedir_context_tpl || !homeroot_context_tpl || !user_context_tpl) {
		retval = STATUS_ERR;
		goto done;
	}

	for (h = homedirs; h; h = h->next) {
		Ustr *temp = ustr_dup_cstr(h->data);

		if (!temp || !ustr_add_cstr(&temp, "/[^/]*")) {
			ustr_sc_free(&temp);
			retval = STATUS_ERR;
			goto done;
		}

		if (write_home_dir_context(out,
					   homedir_context_tpl, FALLBACK_USER,
					   FALLBACK_USER, ustr_cstr(temp),
					   FALLBACK_USER_PREFIX) !=
		    STATUS_SUCCESS) {
			ustr_sc_free(&temp);
			retval = STATUS_ERR;
			goto done;
		}
		if (write_home_root_context(out,
					    homeroot_context_tpl,
					    h->data) != STATUS_SUCCESS) {
			ustr_sc_free(&temp);
			retval = STATUS_ERR;
			goto done;
		}

		ustr_sc_free(&temp);
	}
	if (write_user_context(out, user_context_tpl,
			       ".*", FALLBACK_USER,
			       FALLBACK_USER_PREFIX) != STATUS_SUCCESS) {
		retval = STATUS_ERR;
		goto done;
	}
	if (write_gen_home_dir_context(out, s, user_context_tpl,
				       homedir_context_tpl) != STATUS_SUCCESS) {
		retval = STATUS_ERR;
	}

      done:
	/* Cleanup */
	semanage_list_destroy(&homedirs);
	semanage_list_destroy(&user_context_tpl);
	semanage_list_destroy(&homedir_context_tpl);
	semanage_list_destroy(&homeroot_context_tpl);

	return retval;
}

int semanage_genhomedircon(semanage_handle_t * sh, int usepasswd)
{
	genhomedircon_settings_t s;
	FILE *out = NULL;
	int retval = 0;

	assert(sh);

	s.homedir_template_path =
	    semanage_path(SEMANAGE_TMP, SEMANAGE_HOMEDIR_TMPL);
	s.fcfilepath = semanage_path(SEMANAGE_TMP, SEMANAGE_FC_HOMEDIRS);

	s.usepasswd = usepasswd;
	s.h_semanage = sh;

	if (!(out = fopen(s.fcfilepath, "w"))) {
		/* couldn't open output file */
		ERR(sh, "Could not open the file_context file for writing");
		return STATUS_ERR;
	}

	retval = write_context_file(&s, out);

	fclose(out);
	return retval;
}
